@@ -1,49 +1,58 @@
-use super::LinkedQueue;
-use super::queue::SyncQueue;
-use super::run::Run;
+use {LinkedQueue, Queue, SyncQueue, Run, Task, Delayed, DelayQueue};
+use std::marker::PhantomData;
 use std::sync::{Arc, Mutex, Condvar};
 use std::sync::atomic::{AtomicUsize, Ordering};
-use std::thread;
+use std::{thread, usize};
+use time::Duration;
 
-use self::Lifecycle::*;
-
-pub struct ThreadPool {
-    inner: Arc<ThreadPoolInner>,
+/// A queue that can be used to back a thread pool
+pub trait WorkQueue<T> : SyncQueue<Option<T>> + Clone + Send + 'static {
 }
 
-impl ThreadPool {
-    pub fn fixed_size(size: u32) -> ThreadPool {
-        ThreadPool::new(size, size, LinkedQueue::new())
+impl<T, Q: SyncQueue<Option<T>> + Clone + Send + 'static> WorkQueue<T> for Q {
+}
+
+pub struct ThreadPool<T: Task+'static, Q: WorkQueue<T> = LinkedQueue<Option<T>>> {
+    inner: Arc<ThreadPoolInner<T, Q>>,
+}
+
+impl<T: Task+'static> ThreadPool<T, LinkedQueue<Option<T>>> {
+    pub fn fixed_size(size: u32) -> ThreadPool<T, LinkedQueue<Option<T>>> {
+        ThreadPool::new(size, size, LinkedQueue::with_capacity(usize::MAX))
     }
 
-    pub fn single_thread() -> ThreadPool {
+    pub fn single_thread() -> ThreadPool<T, LinkedQueue<Option<T>>> {
         ThreadPool::fixed_size(1)
     }
+}
 
-    pub fn new<Q>(
-        core_pool_size: u32,
-        maximum_pool_size: u32,
-        work_queue: Q) -> ThreadPool
-            where Q: SyncQueue<Option<Box<Task>>> + Send + Sync + Clone + 'static {
+impl<T: Task+'static, Q: WorkQueue<T>> ThreadPool<T, Q> {
+    pub fn new(core_pool_size: u32,
+               maximum_pool_size: u32,
+               work_queue: Q) -> ThreadPool<T, Q> {
 
         let inner = ThreadPoolInner::new(
             core_pool_size,
             maximum_pool_size,
-            Box::new(work_queue));
+            work_queue);
 
         ThreadPool { inner: Arc::new(inner) }
     }
 
-    pub fn run<F>(&self, task: F) where F: FnOnce() + Send + 'static {
-        self.inner.run(task);
+    pub fn prestart_core_thread(&self) {
+        self.inner.prestart_core_thread();
+    }
+
+    pub fn prestart_all_core_threads(&self) {
+        self.inner.prestart_all_core_threads();
     }
 
     pub fn shutdown(&self) {
-        self.inner.shutdown(SHUTDOWN);
+        self.inner.shutdown(Lifecycle::Shutdown);
     }
 
     pub fn shutdown_now(&self) {
-        self.inner.shutdown(STOP);
+        self.inner.shutdown(Lifecycle::Stop);
     }
 
     pub fn is_shutdown(&self) -> bool {
@@ -55,9 +64,88 @@ impl ThreadPool {
     }
 }
 
-impl Run for ThreadPool {
-    fn run<F>(&self, task: F) where F: FnOnce() + Send + 'static {
-        ThreadPool::run(self, task);
+impl<T: Task+'static, Q: WorkQueue<T>> Run<T> for ThreadPool<T, Q> {
+    fn run(&self, task: T) {
+        self.inner.run(task, true);
+    }
+}
+
+impl<T: Task+'static, Q: WorkQueue<T>> Clone for ThreadPool<T, Q> {
+    fn clone(&self) -> ThreadPool<T, Q> {
+        ThreadPool { inner: self.inner.clone() }
+    }
+}
+
+/// A thread pool that can schedule tasks to run after a given delay, or to
+/// execute periodically. Delayed tasks do not run before their associated
+/// delays, but besides that, there are no real-time guarantees about exactly
+/// when the task will run.
+pub struct ScheduledThreadPool {
+    thread_pool: ThreadPool<Scheduled, DelayQueue<Option<Scheduled>>>,
+}
+
+impl ScheduledThreadPool {
+    pub fn fixed_size(size: u32) -> ScheduledThreadPool {
+        assert!(size > 0, "thread pool size must be greater than 0");
+
+        ScheduledThreadPool {
+            thread_pool: ThreadPool::new(size, size, DelayQueue::new())
+        }
+    }
+
+    pub fn single_thread() -> ScheduledThreadPool {
+        ScheduledThreadPool::fixed_size(1)
+    }
+
+    pub fn schedule_ms<T: Task+'static>(&self, delay: u32, task: T) {
+        let delay = Duration::milliseconds(delay as i64);
+        let task = Scheduled::Delayed(Box::new(task), delay);
+
+        self.thread_pool.inner.run(task, false);
+    }
+}
+
+impl<T: Task+'static> Run<T> for ScheduledThreadPool {
+    fn run(&self, task: T) {
+        self.schedule_ms(0, task);
+    }
+}
+
+impl Clone for ScheduledThreadPool {
+    fn clone(&self) -> ScheduledThreadPool {
+        ScheduledThreadPool { thread_pool: self.thread_pool.clone() }
+    }
+}
+
+enum Scheduled {
+    Delayed(Box<BoxedTask>, Duration),
+}
+
+impl Delayed for Scheduled {
+    fn delay(&self) -> Duration {
+        match *self {
+            Scheduled::Delayed(_, d) => d,
+        }
+    }
+}
+
+impl Task for Scheduled {
+    fn run(self) {
+        debug!("~~ running scheduled task");
+        match self {
+            Scheduled::Delayed(t, _) => t.run_boxed(),
+        }
+    }
+}
+
+trait BoxedTask : Task + 'static {
+    fn run_boxed(self: Box<Self>);
+}
+
+impl<T: Task + 'static> BoxedTask for T {
+    fn run_boxed(self: Box<T>) {
+        let s = *self;
+        s.run();
     }
 }
 
@@ -70,7 +158,7 @@ impl Run for ThreadPool {
 //
 // - Poison the thread pool if something goes critically wrong
 //
-struct ThreadPoolInner {
+struct ThreadPoolInner<T: Task+'static, Q: WorkQueue<T>> {
 
     // Contains the state, condvar, etc..
     core: Arc<Core>,
@@ -80,30 +168,41 @@ struct ThreadPoolInner {
     // None necessarily means that the queue is empty, so rely
     // solely on is_empty to see if the queue is empty (which we must
     // do for example when deciding whether to transition from
-    // SHUTDOWN to TIDYING).  This accommodates special-purpose
-    // queues such as DelayQueues for which poll() is allowed to
-    // return null even if it may later return non-null when delays
-    // expire.
-    work_queue: Box<WorkQueue>,
+    // Shutdown to Tidying).  This accommodates special-purpose
+    // queues such as DelayQueue for which poll() is allowed to return null
+    // even if it may later return non-null when delays expire.
+    work_queue: Q,
+    task: PhantomData<T>,
 }
 
-impl ThreadPoolInner {
+impl<T: Task+'static, Q: WorkQueue<T>> ThreadPoolInner<T, Q> {
 
     fn new(core_pool_size: u32,
            maximum_pool_size: u32,
-           work_queue: Box<WorkQueue>) -> ThreadPoolInner {
+           work_queue: Q) -> ThreadPoolInner<T, Q> {
 
         let core = Arc::new(Core::new(core_pool_size, maximum_pool_size));
 
         ThreadPoolInner {
             core: core,
             work_queue: work_queue,
+            task: PhantomData,
         }
     }
 
-    fn run<F: FnOnce() + Send + 'static>(&self, task: F) {
-        let mut task: Box<Task> = Box::new(task);
+    fn prestart_core_thread(&self) {
+        let _ = self.add_worker(None, true);
+    }
 
+    fn prestart_all_core_threads(&self) {
+        loop {
+            if self.add_worker(None, true).is_err() {
+                return;
+            }
+        }
+    }
+
+    fn run(&self, mut task: T, immediate: bool) {
         //  Proceed in 3 steps:
         //
         //  1. If fewer than `core_pool_size` threads are running, try to
@@ -128,13 +227,20 @@ impl ThreadPoolInner {
         debug!("running task; tp-state={:?}; worker_count={}",
                state.lifecycle(), state.worker_count());
 
+        // An optimization allowing spawning of a worker through with a
+        // specified task
         if state.worker_count() < self.core.core_pool_size {
-            match self.add_worker(Some(task), true) {
-                Ok(_) => {
-                    debug!("worker successfully added; core=true");
-                    return
+            if immediate {
+                match self.add_worker(Some(task), true) {
+                    Ok(_) => {
+                        debug!("worker successfully added; core=true");
+                        return
+                    }
+                    Err(t) => task = t.expect("something went wrong"),
                 }
-                Err(t) => task = t.expect("something went wrong"),
+            } else {
+                // Simply ensure that there are enough running threads
+                let _ = self.add_worker(None, false);
             }
 
             state = self.core.state.load(Ordering::Relaxed);
@@ -171,7 +277,7 @@ impl ThreadPoolInner {
     }
 
     fn shutdown(&self, target: Lifecycle) {
-        // Transition from RUNNING -> SHUTDOWN
+        // Transition from Running -> Shutdown
         let mut state = self.core.state.load(Ordering::Relaxed);
         let mut next;
 
@@ -180,15 +286,15 @@ impl ThreadPoolInner {
 
         loop {
             next = match state.lifecycle() {
-                RUNNING => {
+                Lifecycle::Running=> {
                     if state.worker_count() == 0 {
-                        state.with_lifecycle(TERMINATED)
+                        state.with_lifecycle(Lifecycle::Terminated)
                     } else {
                         state.with_lifecycle(target)
                     }
                 }
-                SHUTDOWN => {
-                    if target == SHUTDOWN {
+                Lifecycle::Shutdown=> {
+                    if target == Lifecycle::Shutdown {
                         return;
                     }
 
@@ -236,8 +342,8 @@ impl ThreadPoolInner {
         self.core.await_termination();
     }
 
-    fn add_worker(&self, task: Option<Box<Task>>, core: bool)
-            -> Result<(), Option<Box<Task>>> {
+    fn add_worker(&self, task: Option<T>, core: bool)
+            -> Result<(), Option<T>> {
 
         // == Transition the state ==
 
@@ -246,14 +352,14 @@ impl ThreadPoolInner {
         'retry: loop {
             let lifecycle = state.lifecycle();
 
-            if lifecycle >= STOP {
-                // If the lifecycle is greater than STOP than never create a
-                // add a new worker
+            if lifecycle >= Lifecycle::Stop {
+                // If the lifecycle is greater than Lifecycle::Stop than never
+                // create a add a new worker
                 return Err(task);
             }
 
-            if lifecycle == SHUTDOWN {
-                // If the lifecycle is currently SHUTDOWN, only add a new
+            if lifecycle == Lifecycle::Shutdown {
+                // If the lifecycle is currently Shutdown, only add a new
                 // worker if it is needed (there is work to process).
                 if task.is_none() && self.work_queue.is_empty() {
                     return Err(task);
@@ -289,7 +395,7 @@ impl ThreadPoolInner {
 
         let mut worker = Worker::new(
             self.core.clone(), task,
-            self.work_queue.boxed_clone());
+            self.work_queue.clone());
 
         debug!("spawning new worker thread");
 
@@ -299,28 +405,32 @@ impl ThreadPoolInner {
     }
 }
 
-impl Drop for ThreadPoolInner {
+impl<T: Task+'static, Q: WorkQueue<T>> Drop for ThreadPoolInner<T, Q> {
     fn drop(&mut self) {
-        self.shutdown(STOP);
+        self.shutdown(Lifecycle::Stop);
     }
 }
 
-struct Worker {
+// Needed because of the PhantomData marker
+unsafe impl<T: Task+'static, Q: WorkQueue<T>> Send for ThreadPoolInner<T, Q> { }
+unsafe impl<T: Task+'static, Q: WorkQueue<T>> Sync for ThreadPoolInner<T, Q> { }
+
+struct Worker<T: Task+'static, Q: WorkQueue<T>> {
     // Core shared by ThreadPool and Worker
     core: Arc<Core>,
 
     // The task to run when the thread first starts
-    initial_task: Option<Box<Task>>,
+    initial_task: Option<T>,
 
     // The queue on which to listen for new tasks
-    work_queue: Box<WorkQueue>,
+    work_queue: Q,
 
     // Checked in the drop function whether or not the thread panicked
     panicked: bool,
 }
 
-impl Worker {
-    fn new(core: Arc<Core>, initial_task: Option<Box<Task>>, queue: Box<WorkQueue>) -> Worker {
+impl<T: Task+'static, Q: WorkQueue<T>> Worker<T, Q> {
+    fn new(core: Arc<Core>, initial_task: Option<T>, queue: Q) -> Worker<T, Q> {
         Worker {
             core: core,
             initial_task: initial_task,
@@ -333,7 +443,8 @@ impl Worker {
         self.panicked = true;
 
         while let Some(task) = self.get_task() {
-            task.invoke();
+            debug!("worker processing task");
+            task.run();
         }
 
         self.panicked = false;
@@ -341,14 +452,14 @@ impl Worker {
 
     // Gets the next task, blocking if necessary. Returns None if the worker
     // should shutdown
-    fn get_task(&mut self) -> Option<Box<Task>> {
+    fn get_task(&mut self) -> Option<T> {
         let mut task = self.initial_task.take();
 
         // Load the state
         let state = self.core.state.load(Ordering::Relaxed);
 
         loop {
-            if state.lifecycle() >= STOP {
+            if state.lifecycle() >= Lifecycle::Stop {
                 debug!("threadpool is stopped -- aborting task get");
 
                 // No more tasks should be removed from the queue, exit the
@@ -370,7 +481,7 @@ impl Worker {
                     debug!("threadpool over max worker count -- shutting thread down; count={}", wc);
 
                     // This can never be a termination state since the
-                    // lifecycle is not STOP or TERMINATE (checked above) and
+                    // lifecycle is not Stop or Terminate (checked above) and
                     // the queue has not been accessed, so it is unknown
                     // whether or not there is a pending task.
                     //
@@ -383,10 +494,11 @@ impl Worker {
                 continue;
             }
 
+            debug!("worker waiting for task");
             match self.work_queue.take() {
                 Some(t) => {
                     // Grab the task, but the loop will restart in order to
-                    // check the state again. If the state transitioned to STOP
+                    // check the state again. If the state transitioned to Stop
                     // while the worker was blocked on the queue, the task
                     // should be discarded and the worker shutdown.
                     task = Some(t);
@@ -411,7 +523,7 @@ impl Worker {
 
         if state.worker_count() == 1 {
             if panicking {
-                if state.lifecycle() >= STOP {
+                if state.lifecycle() >= Lifecycle::Stop {
                     let core = self.core.clone();
 
                     thread::spawn(move || {
@@ -425,7 +537,7 @@ impl Worker {
     }
 }
 
-impl Drop for Worker {
+impl<T: Task+'static, Q: WorkQueue<T>> Drop for Worker<T, Q> {
     fn drop(&mut self) {
         if self.panicked {
             self.decrement_worker_count(true);
@@ -433,18 +545,7 @@ impl Drop for Worker {
     }
 }
 
-trait WorkQueue : SyncQueue<Option<Box<Task>>> + Send + Sync {
-    fn boxed_clone(&self) -> Box<WorkQueue>;
-}
-
-impl<Q: SyncQueue<Option<Box<Task>>> + Send + Sync + Clone + 'static> WorkQueue for Q {
-    fn boxed_clone(&self) -> Box<WorkQueue> {
-        Box::new(self.clone())
-    }
-}
-
 struct Core {
-
     // The main pool control state is an atomic integer packing two conceptual
     // fields
     //   worker_count: indicating the effective number of threads
@@ -463,40 +564,40 @@ struct Core {
     //
     // The `lifecycle` provides the main lifecyle control, taking on values:
     //
-    //   RUNNING:    Accept new tasks and process queued tasks
-    //   SHUTDOWN:   Don't accept new tasks, but process queued tasks
-    //   STOP:       Don't accept new tasks, don't process queued tasks, and
+    //   Running:    Accept new tasks and process queued tasks
+    //   Shutdown:   Don't accept new tasks, but process queued tasks
+    //   Stop:       Don't accept new tasks, don't process queued tasks, and
     //               interrupt in-progress tasks
-    //   TIDYING:    All tasks have terminated, worker_count is zero, the thread
-    //               transitioning to state TIDYING will run the terminated() hook
+    //   Tidying:    All tasks have terminated, worker_count is zero, the thread
+    //               transitioning to state Tidying will run the terminated() hook
     //               method
-    //   TERMINATED: terminated() has completed
+    //   Terminated: terminated() has completed
     //
     // The numerical order among these values matters, to allow ordered
     // comparisons. The lifecycle monotonically increases over time, but need
     // not hit each state. The transitions are:
     //
-    //   RUNNING -> SHUTDOWN
+    //   Running -> Shutdown
     //      On invocation of shutdown(), perhaps implicitly in finalize()
     //
-    //   (RUNNING or SHUTDOWN) -> STOP
+    //   (Running or Shutdown) -> Stop
     //      On invocation of shutdown_now()
     //
-    //   SHUTDOWN -> TIDYING
+    //   Shutdown -> Tidying
     //      When both queue and pool are empty
     //
-    //   STOP -> TIDYING
+    //   Stop -> Tidying
     //      When pool is empty
     //
-    //   TIDYING -> TERMINATED
+    //   Tidying -> Terminated
     //      When the terminated() hook method has completed
     //
     // Threads waiting in await_termination() will return when the state reaches
-    // TERMINATED.
+    // Terminated.
     //
-    // Detecting the transition from SHUTDOWN to TIDYING is less
+    // Detecting the transition from Shutdown to Tidying is less
     // straightforward than you'd like because the queue may become empty after
-    // non-empty and vice versa during SHUTDOWN state, but we can only
+    // non-empty and vice versa during Shutdown state, but we can only
     // terminate if, after seeing that it is empty, we see that workerCount is
     // 0 (which sometimes entails a recheck -- see below).
     state: AtomicState,
@@ -519,7 +620,7 @@ struct Core {
 impl Core {
     fn new(core_pool_size: u32, maximum_pool_size: u32) -> Core {
         Core {
-            state: AtomicState::new(RUNNING),
+            state: AtomicState::new(Lifecycle::Running),
             mutex: Mutex::new(()),
             termination: Condvar::new(),
             core_pool_size: core_pool_size,
@@ -545,7 +646,7 @@ impl Core {
         let _lock = self.mutex.lock()
             .ok().expect("something went wrong");
 
-        // Transition to TERMINATED
+        // Transition to Terminated
         self.state.transition_to_terminated(Ordering::Relaxed);
 
         // Notify all pending threads
@@ -599,7 +700,7 @@ impl AtomicState {
         let mut state = self.load(order);
 
         loop {
-            let next = state.with_lifecycle(TERMINATED);
+            let next = state.with_lifecycle(Lifecycle::Terminated);
             let actual = self.compare_and_swap(state, next, order);
 
             if state == actual {
@@ -639,11 +740,11 @@ impl State {
     }
 
     fn is_running(&self) -> bool {
-        self.lifecycle() == RUNNING
+        self.lifecycle() == Lifecycle::Running
     }
 
     fn is_terminated(&self) -> bool {
-        self.lifecycle() == TERMINATED
+        self.lifecycle() == Lifecycle::Terminated
     }
 
     fn as_u32(&self) -> u32 {
@@ -661,32 +762,24 @@ const CAPACITY: u32 = (1 << (32 - 3)) - 1;
 
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
 enum Lifecycle {
-    RUNNING    = 0,
-    SHUTDOWN   = 1,
-    STOP       = 2,
-    TIDYING    = 3,
-    TERMINATED = 4,
+    Running    = 0,
+    Shutdown   = 1,
+    Stop       = 2,
+    Tidying    = 3,
+    Terminated = 4,
 }
 
 impl Lifecycle {
     fn from_u32(val: u32) -> Lifecycle {
+        use self::Lifecycle::*;
+
         match val {
-            0 => RUNNING,
-            1 => SHUTDOWN,
-            2 => STOP,
-            3 => TIDYING,
-            4 => TERMINATED,
+            0 => Running,
+            1 => Shutdown,
+            2 => Stop,
+            3 => Tidying,
+            4 => Terminated,
             _ => panic!("unexpected state value"),
         }
     }
 }
-
-trait Task : Send {
-    fn invoke(self: Box<Self>);
-}
-
-impl<F: FnOnce() + Send> Task for F {
-    fn invoke(self: Box<F>) {
-        (*self)();
-    }
-}
